"""
Taken from https://github.com/brentp/pyfasta/blob/452d1ce5406ed73c4149b6d201bc65e4aa8afc27/tests/bench.py
"""

from itertools import islice
from tempfile import NamedTemporaryFile
import pyfaidx
import pyfasta
import pysam
from Bio import SeqIO
import time
import random
import os
import sys
from subprocess import call, check_output
import tracemalloc

random.seed(1234)

SEQLEN = 1000000
try:
    nreads = int(sys.argv[1])
except IndexError:
    nreads = 10000
read_len = 1000

def mean(s):
    return sum(s) / len(s)


def make_intervals(nreads=nreads, seqlen=SEQLEN, readlen=read_len):
    for _ in range(nreads):
        start = random.randint(0, seqlen)
        end = min(seqlen, start + readlen)
        yield (start, end)

intervals = tuple(make_intervals())

def make_long_fasta(filename, nrecs=250, seqlen=SEQLEN):
    headers = []
    with open(filename, 'w') as f:
        s = "ACTGACTGAC"
        for i in range(nrecs):
            h = "header%i" % i
            headers.append(h)
            f.write('>' + h + '\n')
            for line in pyfaidx.wrap_sequence(80, s * (seqlen//10)):
                f.write(line)
    return headers

def bgzip_compress_fasta(filename):
    from subprocess import call
    call(' '.join(['bgzip', '-c', filename, '>', filename + '.gz']), shell=True)

def read_dict(f, headers):
    for k in islice(headers, 0, None, 10):
        for start, end in intervals:
            str(f[k][start:end])


def read_faidx(f, headers):
    for k in islice(headers, 0, None, 10):
        for start, end in intervals:
            str(f.fetch(k, start + 1, end))


def read_fastahack(f, headers):
    for k in islice(headers, 0, None, 10):
        for start, end in intervals:
            str(f.get_sub_sequence(k, start, end))


def read_pysam(f, headers):
    tstart = time.time()
    for k in islice(headers, 0, None, 100):
        for start, end in intervals:
            if time.time() - tstart > 300:
                print(k)
            tstart = time.time()
            str(pysam.faidx(f, '{0}:{1}-{2}'.format(k, start + 1, end)))


def read_samtools(f, headers):
    tstart = time.time()
    for k in islice(headers, 0, None, 100):
        for start, end in intervals:
            if time.time() - tstart > 300:
                print(k)
                tstart = time.time()
            check_output(['samtools', 'faidx', f, '{0}:{1}-{2}'.format(k, start + 1, end)])


def main():
    fa_file = NamedTemporaryFile()
    index = fa_file.name + '.fai'
    bgzf_index = fa_file.name + '.gz.fai'
    headers = make_long_fasta(fa_file.name)
    bgzip_compress_fasta(fa_file.name)

    def pyfaidx_fasta(n):
        print('timings for pyfaidx.Fasta')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = pyfaidx.Fasta(fa_file.name)
            ti.append(time.time() - t)

            t = time.time()
            read_dict(f, headers)
            tf.append(time.time() - t)
            os.remove(index)
        # profile memory usage and report timings
        tracemalloc.start()
        f = pyfaidx.Fasta(fa_file.name)
        read_dict(f, headers)
        os.remove(index)
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def pyfaidx_bgzf_faidx(n):
        print('timings for pyfaidx.Faidx with bgzf compression')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = pyfaidx.Faidx(fa_file.name + '.gz')
            ti.append(time.time() - t)

            t = time.time()
            read_faidx(f, headers)
            tf.append(time.time() - t)
            os.remove(index)
        # profile memory usage and report timings
        tracemalloc.start()
        f = pyfaidx.Faidx(fa_file.name + '.gz')
        read_faidx(f, headers)
        os.remove(index)
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def pyfaidx_faidx(n):
        print('timings for pyfaidx.Faidx')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = pyfaidx.Faidx(fa_file.name)
            ti.append(time.time() - t)

            t = time.time()
            read_faidx(f, headers)
            tf.append(time.time() - t)
            os.remove(index)
        # profile memory usage and report timings
        tracemalloc.start()
        f = pyfaidx.Faidx(fa_file.name)
        read_faidx(f, headers)
        os.remove(index)
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def fastahack_fetch(n):
        print('timings for fastahack.FastaHack')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = fastahack.FastaHack(fa_file.name)
            ti.append(time.time() - t)

            t = time.time()
            read_fastahack(f, headers)
            tf.append(time.time() - t)
            os.remove(index)
        # profile memory usage and report timings
        tracemalloc.start()
        f = fastahack.FastaHack(fa_file.name)
        read_fastahack(f, headers)
        os.remove(index)
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def pyfasta_fseek(n):
        print('timings for pyfasta.Fasta (fseek)')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = pyfasta.Fasta(fa_file.name, record_class=pyfasta.FastaRecord)
            ti.append(time.time() - t)

            t = time.time()
            read_dict(f, headers)
            tf.append(time.time() - t)
            os.remove(fa_file.name + '.flat')
            os.remove(fa_file.name + '.gdx')
        # profile memory usage and report timings
        tracemalloc.start()
        f = pyfasta.Fasta(fa_file.name, record_class=pyfasta.FastaRecord)
        read_dict(f, headers)
        os.remove(fa_file.name + '.flat')
        os.remove(fa_file.name + '.gdx')
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def pyfasta_fasta(n):
        print('timings for pyfasta.Fasta')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            f = pyfasta.Fasta(fa_file.name)
            ti.append(time.time() - t)

            t = time.time()
            read_dict(f, headers)
            tf.append(time.time() - t)
            os.remove(fa_file.name + '.flat')
            os.remove(fa_file.name + '.gdx')
        # profile memory usage and report timings
        tracemalloc.start()
        f = pyfasta.Fasta(fa_file.name)
        read_dict(f, headers)
        os.remove(fa_file.name + '.flat')
        os.remove(fa_file.name + '.gdx')
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def pysam_faidx(n):
        print('timings for pysam.faidx')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            pysam.faidx(fa_file.name)
            ti.append(time.time() - t)

            t = time.time()
            read_pysam(fa_file.name, headers)
            tf.append(time.time() - t)
            os.remove(index)
        # profile memory usage and report timings
        tracemalloc.start()
        pysam.faidx(fa_file.name)
        read_pysam(fa_file.name, headers)
        os.remove(index)
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/10*1000*1000)
        tracemalloc.stop()

    def samtools_faidx(n):
        print('timings for samtools faidx')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            call(['samtools', 'faidx', fa_file.name])
            ti.append(time.time() - t)

            t = time.time()
            read_samtools(fa_file.name, headers)
            tf.append(time.time() - t)
            os.remove(index)
        print(mean(ti))
        print(mean(tf)/nreads/100*1000*1000)

    def seqio_read(n):
        print('timings for Bio.SeqIO')
        ti = []
        tf = []
        for _ in range(n):
            t = time.time()
            fh = open(fa_file.name)
            f = SeqIO.to_dict(SeqIO.parse(fh, "fasta"))
            ti.append(time.time() - t)

            t = time.time()
            read_dict(f, headers)
            tf.append(time.time() - t)
            fh.close()
        # profile memory usage and report timings
        tracemalloc.start()
        fh = open(fa_file.name)
        f = SeqIO.to_dict(SeqIO.parse(fh, "fasta"))
        read_dict(f, headers)
        fh.close()
        print(tracemalloc.get_traced_memory())
        print(mean(ti))
        print(mean(tf)/nreads/100*1000*1000)
        tracemalloc.stop()

    n = 3
    pyfaidx_fasta(n)
    pyfaidx_faidx(n)
    pyfaidx_bgzf_faidx(n)
    pyfasta_fasta(n)
    pyfasta_fseek(n)
    seqio_read(n)
    #fastahack_fetch(n)
    samtools_faidx(n)
    pysam_faidx(n)


if __name__ == "__main__":
    main()



import os
import pytest
from pyfaidx import Fasta, FetchError

path = os.path.dirname(__file__)
os.chdir(path)

try:
    from Bio import SeqIO
    bio = True
except ImportError:
    bio = False
    
@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

@pytest.mark.skipif(not bio, reason="Biopython is not installed.")
def test_fetch_whole_entry(remove_index):
    fasta = Fasta('data/genes.fasta')
    with open('data/genes.fasta', "r") as fh:
        seqio = SeqIO.to_dict(SeqIO.parse(fh, "fasta"))
    assert str(fasta['gi|557361099|gb|KF435150.1|']) == str(seqio['gi|557361099|gb|KF435150.1|'].seq)
    assert fasta['gi|557361099|gb|KF435150.1|'].name == str(seqio['gi|557361099|gb|KF435150.1|'].name)

@pytest.mark.skipif(not bio, reason="Biopython is not installed.")
def test_slice_whole_entry(remove_index):
    fasta = Fasta('data/genes.fasta')
    with open('data/genes.fasta', "r") as fh:
        seqio = SeqIO.to_dict(SeqIO.parse(fh, "fasta"))
    assert str(fasta['gi|557361099|gb|KF435150.1|'][::3]) == str(seqio['gi|557361099|gb|KF435150.1|'].seq[::3])

@pytest.mark.skipif(not bio, reason="Biopython is not installed.")
def test_revcomp_whole_entry(remove_index):
    fasta = Fasta('data/genes.fasta')
    with open('data/genes.fasta', "r") as fh:
        seqio = SeqIO.to_dict(SeqIO.parse(fh, "fasta"))
    assert str(fasta['gi|557361099|gb|KF435150.1|'][:].reverse.complement) == str(seqio['gi|557361099|gb|KF435150.1|'].reverse_complement().seq)


import os
import pytest
from pyfaidx import Faidx

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_fetch_border_padded(remove_index):
    """ Fetch past the end of a gene entry """
    faidx = Faidx('data/genes.fasta', default_seq='N')
    expect = 'TCNNNNNNNNNNNNNNNNNNN'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 500)
    assert str(result) == expect



import os
import pytest
from pyfaidx import Fasta
from itertools import chain

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_fetch_whole_fasta(remove_index):
    expect = [line.rstrip('\n') for line in open('data/genes.fasta') if line[0] != '>']
    result = list(chain(*([line for line in record] for record in Fasta('data/genes.fasta', as_raw=True))))
    assert expect == result

def test_line_len(remove_index):
    fasta = Fasta('data/genes.fasta')
    for record in fasta:
        assert len(next(iter(record))) == fasta.faidx.index[record.name].lenc

def test_reverse_iter(remove_index):
    expect = list(chain(*([line[::-1] for line in record][::-1] for record in Fasta('data/genes.fasta', as_raw=True))))
    result = list(chain(*([line for line in reversed(record)] for record in Fasta('data/genes.fasta', as_raw=True))))
    for a, b in zip(expect, result):
        print(a, b)
    assert expect == result



import os
import pytest
from pyfaidx import Faidx, Fasta, FetchError

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_buffer_false(remove_index):
    fasta = Fasta('data/genes.fasta', strict_bounds=True)
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'.lower()
    result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].seq.lower()
    assert result == expect

def test_buffer_true(remove_index):
    fasta = Fasta('data/genes.fasta', read_ahead=300, strict_bounds=True)
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'.lower()
    result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].seq.lower()
    assert result == expect

def test_buffer_exceed(remove_index):
    fasta = Fasta('data/genes.fasta', read_ahead=300, strict_bounds=True)
    expect = 'atgacatcattttccacctctgctcagtgttcaacatctgacagtgcttgcaggatctctcctggacaaatcaatcaggtacgaccaaaactgccgcttttgaagattttgcatgcagcaggtgcgcaaggtgaaatgttcactgttaaagaggtcatgcactatttaggtcagtacataatggtgaagcaactttatgatcagcaggagcagcatatggtatattgtggtggagatcttttgggagaactactgggacgtcagagcttctccgtgaaagacccaagccctctctatgatatgctaagaaagaatcttgtcactttagccactgctactacagcaaagtgcagaggaaagttccacttccagaaaaagaactacagaagacgatatcccc'
    result = fasta['gi|557361099|gb|KF435150.1|'][0:400].seq.lower()
    assert result == expect

def test_bounds_error(remove_index):
    fasta = Fasta('data/genes.fasta', read_ahead=300, strict_bounds=True)
    with pytest.raises(FetchError):
        result = fasta['gi|557361099|gb|KF435150.1|'][100-1:15000].seq.lower()

def test_buffer_value(remove_index):
    with pytest.raises(ValueError):
        Fasta('data/genes.fasta', read_ahead=0.5)



import pytest
from pyfaidx import Sequence, complement

seq = Sequence(name='gi|557361099|gb|KF435150.1|', seq='TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA',
                    start=100, end=150)

seq_invalid = Sequence(name='gi|557361099|gb|KF435150.1|', seq='TTGAAGATTTPGCATGCAGCAGGTGCGCAAGGTGAAATNTTCACTGTTAAA',
                    start=100, end=150)

comp_valid = 'TTGAAGATTTnGCATGCAGCAGGtgccaAGGTGAAATGTTNACTGTTAAA'

comp_invalid = 'TTGAAGATTTnGCATGCAGCPQGtgccaAGGTGAAATGTTNACTGTTAAA'

def test_negate():
    assert str(-seq) == str(seq.complement[::-1])

def test_negate_metadata():
    # Negate should affect __repr__ the same way as reverse and complement
    seq_neg = -seq
    assert seq_neg.__repr__() == seq.complement[::-1].__repr__()

def test_seq_invalid():
    with pytest.raises(ValueError):
        seq_invalid.complement()

def test_integer_index():
    assert seq[1].seq == 'T'

def test_slice_index():
    assert seq[0:10].seq == 'TTGAAGATTT'

def test_comp_invalid():
    with pytest.raises(ValueError):
        complement(comp_invalid)

def test_check_coordinates():
    x = Sequence(name='gi|557361099|gb|KF435150.1|', seq='TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA',
                 start=100, end=110)
    with pytest.raises(ValueError):
        _ = x[:]

def test_comp_valid():
    assert complement(comp_valid).startswith("AACTTCTAAAnCG")
    assert complement(complement(comp_valid)) == comp_valid

def test_comp_empty():
    assert complement('') == ''



import os
import pytest
from pathlib import Path
from pyfaidx import Faidx, Fasta

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_Faidx(remove_index):
    """ Ensures that Faidx can be created with a pathlib.Path as filename """
    filename = 'data/genes.fasta'
    faidx = Faidx(filename)
    faidx_w_path = Faidx(Path(filename))
    assert faidx.filename == faidx_w_path.filename

def test_Fasta(remove_index):
    """ Ensures that Fasta can be created with a pathlib.Path as filename """
    filename = 'data/genes.fasta'
    fasta = Fasta(filename)
    fasta_w_path = Fasta(Path(filename))
    assert fasta.filename == fasta_w_path.filename



import os
import pytest
from pyfaidx import Faidx, Fasta

path = os.path.dirname(__file__)
os.chdir(path)

ACCESSION_TO_GENE_NAME_DICT = {
    'gi|563317589|dbj|AB821309.1|': 'FGFR2',
    'gi|557361099|gb|KF435150.1|': 'MDM4',
    'gi|543583796|ref|NR_104216.1|': 'BARD1',
    # The rest are deliberately omitted
    # KF435149.1, NR_104215.1, NR_104212.1, NM_001282545.1 ...
    }

ACCESSION_TO_DUPLICATED_GENE_NAME_DICT = {
    'gi|563317589|dbj|AB821309.1|': 'FGFR2',
    'gi|557361099|gb|KF435150.1|': 'MDM4',
    'gi|543583796|ref|NR_104216.1|': 'BARD1',
    'gi|543583795|ref|NR_104215.1|': 'BARD1', # Duplicated gene names will trigger a warning
    # The rest are deliberately omitted
    # KF435149.1, NR_104212.1, NM_001282545.1 ...
    }

def get_gene_name(accession):
    '''Return the gene name if found in ACCESSION_TO_GENE_NAME_DICT else return the original accession.'''
    return ACCESSION_TO_GENE_NAME_DICT.get(accession, accession)

def get_duplicated_gene_name(accession):
    '''Return the gene name if found in ACCESSION_TO_GENE_NAME_DICT else return the original accession.'''
    return ACCESSION_TO_DUPLICATED_GENE_NAME_DICT.get(accession, accession)


@pytest.fixture
def remove_index():
    genes = Fasta('data/genes.fasta')
    del genes  # Support feature introduced in #111
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_keys(remove_index):
    genes = Fasta('data/genes.fasta', key_function=get_gene_name)
    expect = ['BARD1', 'FGFR2', 'MDM4', 'gi|530364724|ref|XR_241079.1|', 'gi|530364725|ref|XR_241080.1|', 'gi|530364726|ref|XR_241081.1|', 'gi|530373235|ref|XM_005265507.1|', 'gi|530373237|ref|XM_005265508.1|', 'gi|530384534|ref|XM_005249642.1|', 'gi|530384536|ref|XM_005249643.1|', 'gi|530384538|ref|XM_005249644.1|', 'gi|530384540|ref|XM_005249645.1|', 'gi|543583738|ref|NM_001282548.1|', 'gi|543583740|ref|NM_001282549.1|', 'gi|543583785|ref|NM_000465.3|', 'gi|543583786|ref|NM_001282543.1|', 'gi|543583788|ref|NM_001282545.1|', 'gi|543583794|ref|NR_104212.1|', 'gi|543583795|ref|NR_104215.1|', 'gi|557361097|gb|KF435149.1|']
    result = sorted(genes.keys())
    assert result == expect

def test_key_function_by_dictionary_get_key(remove_index):
    genes = Fasta('data/genes.fasta', key_function=get_gene_name)
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = genes['MDM4'][100-1:150]
    assert str(result) == expect

def test_key_function_by_fetch(remove_index):
    faidx = Faidx('data/genes.fasta', key_function=get_gene_name)
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = faidx.fetch('MDM4',
                         100, 150)
    assert str(result) == expect

def test_duplicated_keys(remove_index):
    with pytest.raises(ValueError):
        genes = Fasta('data/genes.fasta', key_function=get_duplicated_gene_name)

def test_duplicated_keys_shortest(remove_index):
    genes = Fasta('data/genes.fasta', key_function=get_duplicated_gene_name, duplicate_action="shortest")
    expect = 4573
    result = len(genes["BARD1"])
    assert expect == result

def test_duplicated_keys_longest(remove_index):
    genes = Fasta('data/genes.fasta', key_function=get_duplicated_gene_name, duplicate_action="longest")
    expect = 5317
    result = len(genes["BARD1"])
    assert expect == result



#!/usr/bin/env python
import os.path

def fetch_genes(filename, suffix=None):
    from Bio import Entrez
    Entrez.email = "mdshw5@gmail.com"

    id_list = ['563317589', '557361099', '557361097', '543583796',
               '543583795', '543583794', '543583788', '543583786',
               '543583785', '543583740', '543583738', '530384540',
               '530384538', '530384536', '530384534', '530373237',
               '530373235', '530364726', '530364725', '530364724']

    search_results = Entrez.read(Entrez.epost("nucleotide", id=",".join(id_list)))

    webenv = search_results["WebEnv"]
    query_key = search_results["QueryKey"]

    records = Entrez.efetch(db="nucleotide", rettype="fasta", retmode="text", webenv=webenv, query_key=query_key)

    with open(filename, 'w') as fasta:
        for line in records:
            if suffix is not None and line[0] == '>':
                line = line.rstrip('\n')
                line = ''.join([line, suffix, '\n'])
            if len(line) == 1:  # skip lines with only \n
                continue
            fasta.write(line)
    with open(filename, 'r') as fasta:
        with open('.'.join([filename, 'lower']), 'w') as lower:
            for line in fasta:
                if line[0] != '>':
                    line = line.lower()
                lower.write(line)

def fetch_chr22(filename):
    import requests
    import gzip
    import io

    with requests.get('https://ftp-trace.ncbi.nih.gov/1000genomes/ftp/pilot_data/technical/reference/human_b36_male.fa.gz') as compressed:
        with open(filename, 'w') as fasta, gzip.GzipFile(fileobj=io.BytesIO(compressed.content)) as gz:
            chr22 = False
            for line in gz:
                line_content = line.decode()
                if line_content[0:3] == '>22':
                    fasta.write(line_content)
                    chr22 = True
                elif not chr22:
                    continue
                elif chr22 and line_content[0] == '>':
                    break
                elif chr22:
                    fasta.write(line_content)

def add_fake_chr(existing_fasta, filename):
    with open(filename, 'w') as fasta:
        with open(existing_fasta, 'r') as old_fa:
            for line in old_fa:
                fasta.write(line)
        fasta.write('>fake chromosome not in vcf\n')
        fasta.write('ATCG\n')

def fake_chr22(filename):
    """ Fake up some data """
    chr22_len = 49691432
    mod_70 = chr22_len % 70
    with open(filename, 'w') as fake_file:
        fake_file.write('>22 fake data for testing\n')
        while chr22_len > mod_70:
            fake_file.write('N' * 70 + '\n')
            chr22_len -= 70
        fake_file.write('N' * mod_70 + '\n')

def bgzip_compress_fasta(filename):
    from Bio.bgzf import BgzfWriter
    with BgzfWriter(filename=filename + '.gz') as compressed, open(filename, 'r') as fasta:
        for line in fasta:
            compressed.write(line)

def fetch_chr22_vcf(filename):
    import requests
    
    with requests.get('https://ftp-trace.ncbi.nih.gov/1000genomes/ftp/pilot_data/release/2010_07/exon/snps/CEU.exon.2010_03.genotypes.vcf.gz') as vcf:
        with open(filename, 'wb') as out:
            out.write(vcf.content)
    with requests.get('https://ftp-trace.ncbi.nih.gov/1000genomes/ftp/pilot_data/release/2010_07/exon/snps/CEU.exon.2010_03.genotypes.vcf.gz.tbi') as tbi:
        with open(filename + '.tbi', 'wb') as out:
            out.write(tbi.content)


if __name__ == "__main__":
    path = os.path.dirname(__file__)
    os.chdir(path)
    if not os.path.isfile("genes.fasta") or not os.path.isfile("genes.fasta.lower"):
        print("GETTING genes")
        fetch_genes("genes.fasta")
    if not os.path.isfile("chr22.vcf.gz"):
        print("GETTING vcf")
        fetch_chr22_vcf("chr22.vcf.gz")
    if not os.path.isfile("chr22.fasta"):
        print("GETTING chr22.fasta")
        fetch_chr22("chr22.fasta")
    if not os.path.isfile("chr22andfake.fasta"):
        print("adding fake chr")
        add_fake_chr("chr22.fasta", "chr22andfake.fasta")
    bgzip_compress_fasta("genes.fasta")
    bgzip_compress_fasta("chr22.fasta")



import os

import pytest

from pyfaidx import Fasta

try:
    import fsspec
    from fsspec.core import OpenFile
except ImportError:
    pytestmark = pytest.mark.skip


@pytest.fixture(scope="function")
def openfile_genes_fasta():
    testdir = os.path.dirname(__file__)
    genes_fasta = os.path.join(testdir, 'data', 'genes.fasta')

    fs = fsspec.filesystem("memory")
    with fs.open('genes.fasta', mode='wb') as f:
        with open(genes_fasta, mode="rb") as g:
            f.write(g.read())

    try:
        yield fsspec.open('memory://genes.fasta', mode='rb')
    finally:
        fs.rm("/**", recursive=True)
        assert not fs.ls("/")


def test_fsspec_fetch_whole_file(openfile_genes_fasta):
    _ = Fasta(openfile_genes_fasta)


def test_fsspec_default_index(openfile_genes_fasta):
    _ = Fasta(openfile_genes_fasta)

    fs = openfile_genes_fasta.fs
    assert fs.isfile(openfile_genes_fasta.path + ".fai")
    assert fs.size(openfile_genes_fasta.path + ".fai") > 0


def test_fsspec_local_index(openfile_genes_fasta, tmp_path):
    index = tmp_path.joinpath("my_local_index.fai")
    _ = Fasta(openfile_genes_fasta, indexname=index)
    assert index.is_file()
    assert index.stat().st_size > 0


def test_fsspec_remote_index(openfile_genes_fasta):
    f_fai = fsspec.open("memory://some_other_index.fai")
    _ = Fasta(openfile_genes_fasta, indexname=f_fai)

    fs = f_fai.fs
    assert fs.isfile("some_other_index.fai")
    assert fs.size("some_other_index.fai") > 0



import os
import pytest
from os.path import getmtime
from pyfaidx import Faidx, FastaIndexingError, IndexNotFoundError, FastaNotFoundError
from tempfile import NamedTemporaryFile, mkdtemp
import time
import platform
import shutil

try:
    from unittest import mock
except ImportError:
    import mock

import six.moves.builtins as builtins

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_build(remove_index):
    expect_index = ("gi|563317589|dbj|AB821309.1|	3510	114	70	71\n"
                    "gi|557361099|gb|KF435150.1|	481	3789	70	71\n"
                    "gi|557361097|gb|KF435149.1|	642	4368	70	71\n"
                    "gi|543583796|ref|NR_104216.1|	4573	5141	70	71\n"
                    "gi|543583795|ref|NR_104215.1|	5317	9901	70	71\n"
                    "gi|543583794|ref|NR_104212.1|	5374	15415	70	71\n"
                    "gi|543583788|ref|NM_001282545.1|	4170	20980	70	71\n"
                    "gi|543583786|ref|NM_001282543.1|	5466	25324	70	71\n"
                    "gi|543583785|ref|NM_000465.3|	5523	30980	70	71\n"
                    "gi|543583740|ref|NM_001282549.1|	3984	36696	70	71\n"
                    "gi|543583738|ref|NM_001282548.1|	4113	40851	70	71\n"
                    "gi|530384540|ref|XM_005249645.1|	2752	45151	70	71\n"
                    "gi|530384538|ref|XM_005249644.1|	3004	48071	70	71\n"
                    "gi|530384536|ref|XM_005249643.1|	3109	51246	70	71\n"
                    "gi|530384534|ref|XM_005249642.1|	3097	54528	70	71\n"
                    "gi|530373237|ref|XM_005265508.1|	2794	57830	70	71\n"
                    "gi|530373235|ref|XM_005265507.1|	2848	60824	70	71\n"
                    "gi|530364726|ref|XR_241081.1|	1009	63849	70	71\n"
                    "gi|530364725|ref|XR_241080.1|	4884	65009	70	71\n"
                    "gi|530364724|ref|XR_241079.1|	2819	70099	70	71\n")
    index_file = Faidx('data/genes.fasta').indexname
    result_index = open(index_file).read()
    assert result_index == expect_index

def test_build_issue_141(remove_index):
    expect_index = ("gi|563317589|dbj|AB821309.1|	3510	115	70	72\n"
                    "gi|557361099|gb|KF435150.1|	481	3842	70	72\n"
                    "gi|557361097|gb|KF435149.1|	642	4429	70	72\n"
                    "gi|543583796|ref|NR_104216.1|	4573	5213	70	72\n"
                    "gi|543583795|ref|NR_104215.1|	5317	10040	70	72\n"
                    "gi|543583794|ref|NR_104212.1|	5374	15631	70	72\n"
                    "gi|543583788|ref|NM_001282545.1|	4170	21274	70	72\n"
                    "gi|543583786|ref|NM_001282543.1|	5466	25679	70	72\n"
                    "gi|543583785|ref|NM_000465.3|	5523	31415	70	72\n"
                    "gi|543583740|ref|NM_001282549.1|	3984	37211	70	72\n"
                    "gi|543583738|ref|NM_001282548.1|	4113	41424	70	72\n"
                    "gi|530384540|ref|XM_005249645.1|	2752	45784	70	72\n"
                    "gi|530384538|ref|XM_005249644.1|	3004	48745	70	72\n"
                    "gi|530384536|ref|XM_005249643.1|	3109	51964	70	72\n"
                    "gi|530384534|ref|XM_005249642.1|	3097	55292	70	72\n"
                    "gi|530373237|ref|XM_005265508.1|	2794	58640	70	72\n"
                    "gi|530373235|ref|XM_005265507.1|	2848	61675	70	72\n"
                    "gi|530364726|ref|XR_241081.1|	1009	64742	70	72\n"
                    "gi|530364725|ref|XR_241080.1|	4884	65918	70	72\n"
                    "gi|530364724|ref|XR_241079.1|	2819	71079	70	72\n")
    index_file = Faidx('data/issue_141.fasta').indexname
    result_index = open(index_file).read()
    os.remove('data/issue_141.fasta.fai')
    print(result_index)
    assert result_index == expect_index

def test_build_issue_111(remove_index):
    expect_index = ("gi|563317589|dbj|AB821309	3510	114	70	71\n"
                    "gi|557361099|gb|KF435150	481	3789	70	71\n"
                    "gi|557361097|gb|KF435149	642	4368	70	71\n"
                    "gi|543583796|ref|NR_104216	4573	5141	70	71\n"
                    "gi|543583795|ref|NR_104215	5317	9901	70	71\n"
                    "gi|543583794|ref|NR_104212	5374	15415	70	71\n"
                    "gi|543583788|ref|NM_001282545	4170	20980	70	71\n"
                    "gi|543583786|ref|NM_001282543	5466	25324	70	71\n"
                    "gi|543583785|ref|NM_000465	5523	30980	70	71\n"
                    "gi|543583740|ref|NM_001282549	3984	36696	70	71\n"
                    "gi|543583738|ref|NM_001282548	4113	40851	70	71\n"
                    "gi|530384540|ref|XM_005249645	2752	45151	70	71\n"
                    "gi|530384538|ref|XM_005249644	3004	48071	70	71\n"
                    "gi|530384536|ref|XM_005249643	3109	51246	70	71\n"
                    "gi|530384534|ref|XM_005249642	3097	54528	70	71\n"
                    "gi|530373237|ref|XM_005265508	2794	57830	70	71\n"
                    "gi|530373235|ref|XM_005265507	2848	60824	70	71\n"
                    "gi|530364726|ref|XR_241081	1009	63849	70	71\n"
                    "gi|530364725|ref|XR_241080	4884	65009	70	71\n"
                    "gi|530364724|ref|XR_241079	2819	70099	70	71\n")
    index = Faidx(
        'data/genes.fasta',
        read_long_names=True,
        key_function=lambda x: x.split('.')[0])
    result_index = ''.join(index._index_as_string())
    assert result_index == expect_index

def test_order(remove_index):
    order = ("gi|563317589|dbj|AB821309.1|", "gi|557361099|gb|KF435150.1|",
             "gi|557361097|gb|KF435149.1|",
             "gi|543583796|ref|NR_104216.1|",
             "gi|543583795|ref|NR_104215.1|",
             "gi|543583794|ref|NR_104212.1|",
             "gi|543583788|ref|NM_001282545.1|",
             "gi|543583786|ref|NM_001282543.1|",
             "gi|543583785|ref|NM_000465.3|",
             "gi|543583740|ref|NM_001282549.1|",
             "gi|543583738|ref|NM_001282548.1|",
             "gi|530384540|ref|XM_005249645.1|",
             "gi|530384538|ref|XM_005249644.1|",
             "gi|530384536|ref|XM_005249643.1|",
             "gi|530384534|ref|XM_005249642.1|",
             "gi|530373237|ref|XM_005265508.1|",
             "gi|530373235|ref|XM_005265507.1|",
             "gi|530364726|ref|XR_241081.1|",
             "gi|530364725|ref|XR_241080.1|",
             "gi|530364724|ref|XR_241079.1|")
    result = tuple(Faidx('data/genes.fasta').index.keys())
    assert result == order

def test_valgrind_short_lines(remove_index):
    """ Makes all full-length lines short and checks that error is raised
    in all appropriate circumstances.
    """
    # http://stackoverflow.com/a/23212515/717419
    if platform.system() == 'Windows':
        raise SkipTest
    indexed = []
    with open('data/genes.fasta') as genes:
        fasta = genes.readlines()
    n_lines = sum(1 for line in fasta)
    for n in range(n_lines):
        with NamedTemporaryFile(mode='w') as lines:
            for i, line in enumerate(fasta):
                if i == n and line[0] != '>' and len(line) == 71:
                    line = line[:-3] + '\n'
                    full_line = True
                elif i == n:
                    full_line = False
                lines.write(line)
                lines.flush()
            name = lines.name
            if full_line:
                try:
                    Faidx(name)
                    indexed.append(True)
                except FastaIndexingError:
                    indexed.append(False)
    assert not any(indexed)

def test_valgrind_long_lines(remove_index):
    """ Makes all full-length lines long and checks that error is raised
    in all appropriate circumstances.
    """
    # http://stackoverflow.com/a/23212515/717419
    if platform.system() == 'Windows':
        raise SkipTest
    indexed = []
    with open('data/genes.fasta') as genes:
        fasta = genes.readlines()
    n_lines = sum(1 for line in fasta)
    for n in range(n_lines):
        with NamedTemporaryFile(mode='w') as lines:
            for i, line in enumerate(fasta):
                if i == n and line[0] != '>' and len(line) == 71:
                    line = line.rstrip('\n') + 'NNN' + '\n'
                    full_line = True
                elif i == n:
                    full_line = False
                lines.write(line)
                lines.flush()
            name = lines.name
            if full_line:
                try:
                    Faidx(name)
                    indexed.append(True)
                except FastaIndexingError:
                    indexed.append(False)
    assert not any(indexed)

def test_valgrind_blank_lines(remove_index):
    """ Makes all full-length lines blank and checks that error is raised
    in all appropriate circumstances.
    """
    # http://stackoverflow.com/a/23212515/717419
    if platform.system() == 'Windows':
        raise SkipTest
    indexed = []
    with open('data/genes.fasta') as genes:
        fasta = genes.readlines()
    n_lines = sum(1 for line in fasta)
    for n in range(n_lines):
        with NamedTemporaryFile(mode='w') as lines:
            for i, line in enumerate(fasta):
                if i == n and line[0] != '>' and len(line) == 71:
                    line = '\n'
                    full_line = True
                elif i == n:
                    full_line = False
                lines.write(line)
                lines.flush()
            name = lines.name
            if full_line:
                try:
                    Faidx(name)
                    indexed.append(True)
                except FastaIndexingError:
                    indexed.append(False)
    assert not any(indexed)

def test_reindex_on_modification(remove_index):
    """ This test ensures that the index is regenerated when the FASTA
    modification time is newer than the index modification time.
    mdshw5/pyfaidx#50 """
    faidx = Faidx('data/genes.fasta')
    index_mtime = getmtime(faidx.indexname)
    faidx.close()
    os.utime('data/genes.fasta', (index_mtime + 10, ) * 2)
    time.sleep(2)
    faidx = Faidx('data/genes.fasta')
    assert getmtime(faidx.indexname) > index_mtime

def test_build_issue_83(remove_index):
    """ Ensure that blank lines between entries are treated in the
    same way as samtools 1.2. See mdshw5/pyfaidx#83.
    """
    expect_index = ("MT	119	4	70	71\nGL000207.1	60	187	60	61\n")
    index_file = Faidx('data/issue_83.fasta').indexname
    result_index = open(index_file).read()
    os.remove('data/issue_83.fasta.fai')
    assert result_index == expect_index

def test_build_issue_96_fail_build_faidx(remove_index):
    """ Ensure that the fasta file is closed if construction of the 'Faidx' file
    when attempting to build an index.
    See mdshw5/pyfaidx#96
    """
    tmp_dir = mkdtemp()
    try:
        fasta_path = os.path.join(tmp_dir, 'issue_96.fasta')
        # Write simple fasta file with inconsistent sequence line lengths,
        # so building an index raises a 'FastaIndexingError'
        with open(fasta_path, 'w') as fasta_out:
            fasta_out.write(
                ">seq1\nCTCCGGGCCCAT\nAACACTTGGGGGTAGCTAAAGTGAA\nATAAAGCCTAAA\n"
            )

        builtins_open = builtins.open

        opened_files = []

        def test_open(*args, **kwargs):
            f = builtins_open(*args, **kwargs)
            opened_files.append(f)
            return f

        with mock.patch('six.moves.builtins.open', side_effect=test_open):
            try:
                Faidx(fasta_path)
                remove_index.assertFail(
                    "Faidx construction should fail with 'FastaIndexingError'."
                )
            except FastaIndexingError:
                pass
        assert all(f.closed for f in opened_files)
    finally:
        shutil.rmtree(tmp_dir)

def test_build_issue_96_fail_read_malformed_index_duplicate_key(remove_index):
    """ Ensure that the fasta file is closed if construction of the 'Faidx' file
    fails when attempting to read a pre-existing index. The index is malformed because
    it contains mulitple occurrences of the same index.
    See mdshw5/pyfaidx#96
    """
    tmp_dir = mkdtemp()
    try:
        fasta_path = os.path.join(tmp_dir, 'issue_96.fasta')
        faidx_path = os.path.join(tmp_dir, 'issue_96.fasta.fai')
        # Write simple fasta file
        with open(fasta_path, 'w') as fasta_out:
            fasta_out.write(">seq1\nCTCCGGGCCCAT\nATAAAGCCTAAA\n")
        with open(faidx_path, 'w') as faidx_out:
            faidx_out.write("seq1\t24\t6\t12\t13\nseq1\t24\t6\t12\t13\n")

        builtins_open = builtins.open

        opened_files = []

        def test_open(*args, **kwargs):
            f = builtins_open(*args, **kwargs)
            opened_files.append(f)
            return f

        with mock.patch('six.moves.builtins.open', side_effect=test_open):
            try:
                Faidx(fasta_path)
                remove_index.assertFail(
                    "Faidx construction should fail with 'ValueError'.")
            except ValueError:
                pass
        assert all(f.closed for f in opened_files)
    finally:
        shutil.rmtree(tmp_dir)

def test_read_back_index(remove_index):
    """Ensure that index files written with write_fai() can be read back"""
    import locale
    import platform
    
    if platform.system() == "Linux":
        new_locale = 'en_US.utf8'
    elif platform.system() == "Darwin":
        new_locale = 'en_US.UTF-8'
    
    old_locale = locale.getlocale(locale.LC_NUMERIC)
    try:
        locale.setlocale(locale.LC_NUMERIC, new_locale)
        faidx = Faidx('data/genes.fasta')
        faidx.write_fai()
        faidx = Faidx('data/genes.fasta', build_index=False)
    finally:
        locale.setlocale(locale.LC_NUMERIC, old_locale)

def test_issue_134_no_build_index(remove_index):
    """ Ensure that index file is not built when build_index=False. See mdshw5/pyfaidx#134.
    """
    with pytest.raises(IndexNotFoundError):
        faidx = Faidx('data/genes.fasta', build_index=False)

def test_issue_144_no_defline(remove_index):
    """ Ensure that an exception is raised when a file contains no deflines. See mdshw5/pyfaidx#144.
    """
    tmp_dir = mkdtemp()
    try:
        fasta_path = os.path.join(tmp_dir, 'issue_144.fasta')
        # Write simple fasta file
        with open(fasta_path, 'w') as fasta_out:
            fasta_out.write("CTCCGGGCCCAT\nATAAAGCCTAAA\n")
        with pytest.raises(FastaIndexingError):
            faidx = Faidx(fasta_path)
    finally:
        shutil.rmtree(tmp_dir)


import os
import pytest
from pyfaidx import FastaVariant, Fasta

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/chr22.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_fetch_variant(remove_index):
    try:
        import pysam
        fasta = FastaVariant('data/chr22.fasta', 'data/chr22.vcf.gz', hom=True, het=True, as_raw=True)
        assert fasta['22'][32330458:32330462] == 'CAGG'  # het
        assert fasta['22'][32352282:32352286] == 'CAGC'  # hom
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")

def test_fetch_hom_variant(remove_index):
    try:
        import pysam
        fasta = FastaVariant('data/chr22.fasta', 'data/chr22.vcf.gz', hom=True, het=False, as_raw=True)
        assert fasta['22'][32330458:32330462] == 'CGGG'  # het
        assert fasta['22'][32352282:32352286] == 'CAGC'  # hom
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")

def test_fetch_het_variant(remove_index):
    try:
        import pysam
        fasta = FastaVariant('data/chr22.fasta', 'data/chr22.vcf.gz', hom=False, het=True, as_raw=True)
        assert fasta['22'][32330458:32330462] == 'CAGG'  # het
        assert fasta['22'][32352282:32352286] == 'CGGC'  # hom
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")

def test_fetch_chr_not_in_vcf(remove_index):
    try:
        import pysam
        fasta = FastaVariant('data/chr22andfake.fasta', 'data/chr22.vcf.gz', hom=True, het=True, as_raw=True)
        assert fasta['fake'][:10] == 'ATCG' # fake is not in vcf 
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")
    
def test_all_pos(remove_index):
    try:
        import pysam
        fasta = FastaVariant('data/chr22.fasta', 'data/chr22.vcf.gz', hom=True, het=True, as_raw=True)
        assert fasta['22'].variant_sites == (16042793, 21833121, 29153196, 29187373, 29187448, 29194610, 29821295, 29821332, 29993842, 32330460, 32352284)
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")

def test_all_diff(remove_index):
    try:
        fasta = FastaVariant('data/chr22.fasta', 'data/chr22.vcf.gz', hom=True, het=True, as_raw=True)
        ref = Fasta('data/chr22.fasta', as_raw=True)
        print([(ref['22'][pos-1], fasta['22'][pos-1]) for pos in fasta['22'].variant_sites])
        assert all(ref['22'][pos-1] != fasta['22'][pos-1] for pos in fasta['22'].variant_sites)
    except (ImportError, IOError):
        pytest.skip("pysam not installed.")



import os
import pytest
from pyfaidx import Faidx, Fasta

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_as_raw_false(remove_index):
    fasta = Fasta('data/genes.fasta')
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'.lower()
    result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].seq.lower()
    assert result == expect

def test_as_raw_true(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True)
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'.lower()
    result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].lower()
    assert result == expect

def test_as_raw_false_error(remove_index):
    fasta = Fasta('data/genes.fasta')
    with pytest.raises(AttributeError):
        result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].lower()

def test_as_raw_true_error(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True)
    with pytest.raises(AttributeError):
        result = fasta['gi|557361099|gb|KF435150.1|'][100-1:150].seq.lower()

def test_as_raw_type_when_blen_lt_0(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True)
    expect = ''
    result = fasta.faidx.fetch('gi|557361099|gb|KF435150.1|', 10, 0)
    assert result == expect



import os
import pytest
from pyfaidx import Fasta

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    fais = [
            "data/gene.bed12.fasta.fai",
            "data/chr17.hg19.part.fa.fai"
            ]
    for fai in fais:
        try:    
            os.remove(fai)
        except EnvironmentError:
            pass  # some tests may delete this file
    
def test_split_seq(remove_index):
    """ Fetch sequence by blocks """
    fa = Fasta('data/chr17.hg19.part.fa')
    
    gene = Fasta("data/gene.bed12.fasta")
    expect = gene[list(gene.keys())[0]][:].seq
    
    bed = "data/gene.bed12"
    with open(bed) as fi:
        record = fi.readline().strip().split("\t")

    chrom = record[0]
    start = int(record[1])
    strand = record[5]

    # parse bed12 format
    starts = [int(x) for x in record[11].split(",")[:-1]] 
    sizes = [int(x) for x in record[10].split(",")[:-1]]
    starts = [start + x  for x in starts]
    ends = [start + size  for start,size in zip(starts, sizes)] 
    
    # bed half-open
    if strand == "-":
        starts = [start + 1 for start in starts]
    else: 
        ends = [end - 1 for end in ends]
    
    intervals = zip(starts, ends) 
    result = fa.get_spliced_seq(chrom, intervals, rc=True)
    print(result.seq)
    print("====")
    print(expect)

    assert result.seq == expect

def test_get_seq_rc(remove_index):
    """ Check get_seq with rc argument """
    fa = Fasta('data/chr17.hg19.part.fa')
    
    result = fa.get_seq("chr17", 11, 20, rc=False)
    expect = "CCCTGTTCCT"
    print("normal")
    print(result.seq)
    print(expect)
    assert result.seq == expect
    
    result = fa.get_seq("chr17", 11, 20, rc=True)
    expect = "AGGAACAGGG"
    assert result.seq == expect
    print("rc")
    print(result.seq)
    print(expect)



import os
import pytest
from pyfaidx import Fasta, Faidx, UnsupportedCompressionFormat, FetchError
from itertools import chain

path = os.path.dirname(__file__)
os.chdir(path)

try:
    from Bio import SeqIO
    bio = True
except ImportError:
    bio = False
    
@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.gz.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

@pytest.mark.skipif(not bio, reason="Biopython is not installed.")
@pytest.mark.xfail
def test_build_issue_126(remove_index):
    """ Samtools BGZF index should be identical to pyfaidx BGZF index """
    expect_index = ("gi|563317589|dbj|AB821309.1|	3510	114	70	71\n"
                    "gi|557361099|gb|KF435150.1|	481	3789	70	71\n"
                    "gi|557361097|gb|KF435149.1|	642	4368	70	71\n"
                    "gi|543583796|ref|NR_104216.1|	4573	5141	70	71\n"
                    "gi|543583795|ref|NR_104215.1|	5317	9901	70	71\n"
                    "gi|543583794|ref|NR_104212.1|	5374	15415	70	71\n"
                    "gi|543583788|ref|NM_001282545.1|	4170	20980	70	71\n"
                    "gi|543583786|ref|NM_001282543.1|	5466	25324	70	71\n"
                    "gi|543583785|ref|NM_000465.3|	5523	30980	70	71\n"
                    "gi|543583740|ref|NM_001282549.1|	3984	36696	70	71\n"
                    "gi|543583738|ref|NM_001282548.1|	4113	40851	70	71\n"
                    "gi|530384540|ref|XM_005249645.1|	2752	45151	70	71\n"
                    "gi|530384538|ref|XM_005249644.1|	3004	48071	70	71\n"
                    "gi|530384536|ref|XM_005249643.1|	3109	51246	70	71\n"
                    "gi|530384534|ref|XM_005249642.1|	3097	54528	70	71\n"
                    "gi|530373237|ref|XM_005265508.1|	2794	57830	70	71\n"
                    "gi|530373235|ref|XM_005265507.1|	2848	60824	70	71\n"
                    "gi|530364726|ref|XR_241081.1|	1009	63849	70	71\n"
                    "gi|530364725|ref|XR_241080.1|	4884	65009	70	71\n"
                    "gi|530364724|ref|XR_241079.1|	2819	70099	70	71\n")
    index_file = Faidx('data/genes.fasta.gz').indexname
    result_index = open(index_file).read()
    assert result_index == expect_index

def test_integer_slice(remove_index):
    fasta = Fasta('data/genes.fasta.gz')
    expect = fasta['gi|563317589|dbj|AB821309.1|'][:100].seq
    result = fasta[0][:100].seq
    assert expect == result

def test_integer_index(remove_index):
    fasta = Fasta('data/genes.fasta.gz')
    expect = fasta['gi|563317589|dbj|AB821309.1|'][100].seq
    result = fasta[0][100].seq
    assert expect == result

def test_fetch_whole_fasta(remove_index):
    expect = [line.rstrip('\n') for line in open('data/genes.fasta') if line[0] != '>']
    result = list(chain(*([line for line in record] for record in Fasta('data/genes.fasta.gz', as_raw=True))))
    assert expect == result

def test_line_len(remove_index):
    fasta = Fasta('data/genes.fasta.gz')
    for record in fasta:
        assert len(next(iter(record))) == fasta.faidx.index[record.name].lenc

def test_mutable_bgzf(remove_index):
    with pytest.raises(UnsupportedCompressionFormat):
        fasta = Fasta('data/genes.fasta.gz', mutable=True)

@pytest.mark.xfail(raises=NotImplementedError)
def test_long_names(remove_index):
    """ Test that deflines extracted using FastaRecord.long_name are
    identical to deflines in the actual file.
    """
    deflines = []
    with open('data/genes.fasta') as fasta_file:
        for line in fasta_file:
            if line[0] == '>':
                deflines.append(line[1:-1])
    fasta = Fasta('data/genes.fasta.gz')
    long_names = []
    for record in fasta:
        long_names.append(record.long_name)
    assert deflines == long_names

def test_fetch_whole_entry(remove_index):
    faidx = Faidx('data/genes.fasta.gz')
    expect = ('ATGACATCATTTTCCACCTCTGCTCAGTGTTCAACATCTGA'
            'CAGTGCTTGCAGGATCTCTCCTGGACAAATCAATCAGGTACGACCA'
            'AAACTGCCGCTTTTGAAGATTTTGCATGCAGCAGGTGCGCAAGG'
            'TGAAATGTTCACTGTTAAAGAGGTCATGCACTATTTAGGTCAGTACAT'
            'AATGGTGAAGCAACTTTATGATCAGCAGGAGCAGCATATGGTATATTG'
            'TGGTGGAGATCTTTTGGGAGAACTACTGGGACGTCAGAGCTTCTCCGTG'
            'AAAGACCCAAGCCCTCTCTATGATATGCTAAGAAAGAATCTTGTCACTTT'
            'AGCCACTGCTACTACAGCAAAGTGCAGAGGAAAGTTCCACTTCCAGAAAAA'
            'GAACTACAGAAGACGATATCCCCACACTGCCTACCTCAGAGCATAAATGCA'
            'TACATTCTAGAGAAGGTGATTGAAGTGGGAAAAAATGATGACCTGGAGGACTC')
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         1, 481)
    assert str(result) == expect

def test_fetch_middle(remove_index):
    faidx = Faidx('data/genes.fasta.gz')
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         100, 150)
    assert str(result) == expect

def test_fetch_end(remove_index):
    faidx = Faidx('data/genes.fasta.gz')
    expect = 'TC'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 481)
    assert str(result) == expect

def test_fetch_border(remove_index):
    """ Fetch past the end of a gene entry """
    faidx = Faidx('data/genes.fasta.gz')
    expect = 'TC'
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', 480, 500)
        print(result)
        assert str(result) == expect

def test_rev(remove_index):
    faidx = Faidx('data/genes.fasta.gz')
    expect = 'GA'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 481)
    assert str(-result) == expect, result

def test_fetch_past_bounds(remove_index):
    """ Fetch past the end of a gene entry """
    faidx = Faidx('data/genes.fasta.gz', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', 480, 5000)

def test_fetch_negative(remove_index):
    """ Fetch starting with a negative coordinate """
    faidx = Faidx('data/genes.fasta.gz', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', -10, 10)

def test_fetch_reversed_coordinates(remove_index):
    """ Fetch starting with a negative coordinate """
    faidx = Faidx('data/genes.fasta.gz', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', 50, 10)

def test_fetch_keyerror(remove_index):
    """ Fetch a key that does not exist """
    faidx = Faidx('data/genes.fasta.gz', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|joe|gb|KF435150.1|', 1, 10)

def test_blank_string(remove_index):
    """ seq[0:0] should return a blank string mdshw5/pyfaidx#53 """
    fasta = Fasta('data/genes.fasta.gz', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][0:0] == ''

def test_slice_from_beginning(remove_index):
    fasta = Fasta('data/genes.fasta.gz', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][:4] == 'ATGA'

def test_slice_from_end(remove_index):
    fasta = Fasta('data/genes.fasta.gz', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][-4:] == 'ACTC'

def test_issue_74_start(remove_index):
    f0 = Fasta('data/genes.fasta.gz', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta.gz', one_based_attributes=True)
    assert f0['gi|557361099|gb|KF435150.1|'][0:90].start == f1['gi|557361099|gb|KF435150.1|'][0:90].start - 1

def test_issue_74_consistency(remove_index):
    f0 = Fasta('data/genes.fasta.gz', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta.gz', one_based_attributes=True)
    assert str(f0['gi|557361099|gb|KF435150.1|'][0:90]) == str(f1['gi|557361099|gb|KF435150.1|'][0:90])

def test_issue_74_end_faidx(remove_index):
    f0 = Faidx('data/genes.fasta.gz', one_based_attributes=False)
    f1 = Faidx('data/genes.fasta.gz', one_based_attributes=True)
    end0 = f0.fetch('gi|557361099|gb|KF435150.1|', 1, 90).end
    end1 = f1.fetch('gi|557361099|gb|KF435150.1|', 1, 90).end
    assert end0 == end1

def test_issue_74_end_fasta(remove_index):
    f0 = Fasta('data/genes.fasta.gz', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta.gz', one_based_attributes=True)
    end0 = f0['gi|557361099|gb|KF435150.1|'][1:90].end
    end1 = f1['gi|557361099|gb|KF435150.1|'][1:90].end
    print((end0, end1))
    assert end0 == end1

def test_issue_79_fix(remove_index):
    f = Fasta('data/genes.fasta.gz')
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print((s.start, s.end))
    assert (101, 105) == (s.start, s.end)

def test_issue_79_fix_negate(remove_index):
    f = Fasta('data/genes.fasta.gz')
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    s = -s
    print((s.start, s.end))
    assert (105, 101) == (s.start, s.end)

def test_issue_79_fix_one_based_false(remove_index):
    f = Fasta('data/genes.fasta.gz', one_based_attributes=False)
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print((s.start, s.end))
    assert (100, 105) == (s.start, s.end)

def test_issue_79_fix_one_based_false_negate(remove_index):
    f = Fasta('data/genes.fasta.gz', one_based_attributes=False)
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print(s.__dict__)
    s = -s
    print(s.__dict__)
    assert (105, 100) == (s.start, s.end)

def test_fetch_border_padded(remove_index):
    """ Fetch past the end of a gene entry """
    with pytest.raises(FetchError):
        faidx = Faidx('data/genes.fasta.gz', default_seq='N')
        expect = 'TCNNNNNNNNNNNNNNNNNNN'
        result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                             480, 500)
        print(result)
        assert str(result) == expect



import os
import pytest
from pyfaidx import Faidx, Fasta

path = os.path.dirname(__file__)
os.chdir(path)


@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file
    
def test_keys(remove_index):
    fasta = Fasta('data/genes.fasta', split_char='|', duplicate_action="drop")
    expect = ['530364724', '530364725', '530364726', '530373235', '530373237', '530384534', '530384536', '530384538', '530384540', '543583738', '543583740', '543583785', '543583786', '543583788', '543583794', '543583795', '543583796', '557361097', '557361099', '563317589', 'AB821309.1', 'KF435149.1', 'KF435150.1', 'NM_000465.3', 'NM_001282543.1', 'NM_001282545.1', 'NM_001282548.1', 'NM_001282549.1', 'NR_104212.1', 'NR_104215.1', 'NR_104216.1', 'XM_005249642.1', 'XM_005249643.1', 'XM_005249644.1', 'XM_005249645.1', 'XM_005265507.1', 'XM_005265508.1', 'XR_241079.1', 'XR_241080.1', 'XR_241081.1', 'dbj']
    result = sorted(fasta.keys())
    assert result == expect

def test_key_function_by_dictionary_get_key(remove_index):
    fasta = Fasta('data/genes.fasta', split_char='|', duplicate_action="drop")
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = fasta['KF435150.1'][100-1:150]
    assert str(result) == expect

def test_key_function_by_fetch(remove_index):
    faidx = Faidx('data/genes.fasta', split_char='|', duplicate_action="drop")
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = faidx.fetch('KF435150.1',
                         100, 150)
    assert str(result) == expect

def test_stop(remove_index):
    with pytest.raises(ValueError):
        fasta = Fasta('data/genes.fasta', split_char='|')



import os
import filecmp
import pytest
from pyfaidx import BedError, FetchError
from pyfaidx.cli import main
from tempfile import NamedTemporaryFile

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

def test_short_line_lengths(remove_index):
    with pytest.raises(BedError):
        main(['data/genes.fasta', '--bed', 'data/malformed.bed'])

def test_fetch_whole_file(remove_index):
    main(['data/genes.fasta'])

def test_split_entry(remove_index):
    main(['--split-files', 'data/genes.fasta', 'gi|557361099|gb|KF435150.1|'])
    assert os.path.exists('gi557361099gbKF435150.1.fasta')
    os.remove('gi557361099gbKF435150.1.fasta')

def test_fetch_error(remove_index):
    with pytest.raises(FetchError):
        main(['data/genes.fasta', 'gi|557361099|gb|KF435150.1|:1-1000'])
    
def test_key_warning(remove_index):
    main(['data/genes.fasta', 'foo'])
    
def test_auto_strand(remove_index):
    """ Test that --auto-strand produces the same output as --reverse --complement"""
    with NamedTemporaryFile() as auto_strand:
        with NamedTemporaryFile() as noto_strand:
            main(['--auto-strand', '-o', auto_strand.name, 'data/genes.fasta', 'gi|557361099|gb|KF435150.1|:100-1'])
            main(['--reverse', '--complement', '-o', noto_strand.name, 'data/genes.fasta', 'gi|557361099|gb|KF435150.1|:1-100'])
            print(auto_strand.read())
            print()
            print(noto_strand.read())
            assert filecmp.cmp(auto_strand.name, noto_strand.name)
    
def test_regexp(remove_index):
    main(['data/genes.fasta', '-g', 'XR'])

def test_not_regexp(remove_index):
    main(['data/genes.fasta', '-g', 'XR','-v'])

def test_not_regexp_multi(remove_index):
    main(['data/genes.fasta', '-g', 'XR', '-g', 'XM', '-v'])



import os
import sys
import pytest
from pyfaidx import Fasta
from tempfile import NamedTemporaryFile
from difflib import Differ

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file
    
def test_sequence_uppercase(remove_index):
    """Test that the sequence is always returned in
    uppercase, even if it is in lowercase in the
    reference genome.
    """
    filename = "data/genes.fasta.lower"
    reference_upper = Fasta(filename, sequence_always_upper=True)
    reference_normal = Fasta(filename)
    os.remove('data/genes.fasta.lower.fai')
    assert reference_upper['gi|557361099|gb|KF435150.1|'][
        1:100].seq == reference_normal['gi|557361099|gb|KF435150.1|'][
            1:100].seq.upper()

def test_long_names(remove_index):
    """ Test that deflines extracted using FastaRecord.long_name are
    identical to deflines in the actual file.
    """
    deflines = []
    with open('data/genes.fasta') as fasta_file:
        for line in fasta_file:
            if line[0] == '>':
                deflines.append(line[1:-1])
    fasta = Fasta('data/genes.fasta')
    long_names = []
    for record in fasta:
        long_names.append(record.long_name)
    print(tuple(zip(deflines, long_names)))
    assert deflines == long_names

def test_issue_62(remove_index):
    """ Check for pathogenic FastaRecord.long_name behavior in mdshw5/pyfaidx#62 """
    deflines = []
    line_len = None
    with open('data/genes.fasta', 'rb') as fasta_file:
        with open('data/issue_62.fa', 'wb') as fasta_uniform_len:
            for line in fasta_file:
                if line.startswith(b'>'):
                    deflines.append(line[1:-1].decode('ascii'))
                    fasta_uniform_len.write(line)
                elif line_len is None:
                    line_len = len(line)
                    fasta_uniform_len.write(line)
                elif line_len > len(line):
                    fasta_uniform_len.write(line.rstrip() + b'N' *
                                            (line_len - len(line)) + b'\n')
                else:
                    fasta_uniform_len.write(line)
    fasta = Fasta('data/issue_62.fa', as_raw=True)
    long_names = []
    for record in fasta:
        long_names.append(record.long_name)
    try:
        os.remove('data/issue_62.fa')
        os.remove('data/issue_62.fa.fai')
    except EnvironmentError:
        pass
    sys.stdout.writelines(tuple(Differ().compare(deflines, long_names)))
    assert deflines == long_names

def test_unpadded_length(remove_index):
    filename = "data/padded.fasta"
    with open(filename, 'w') as padded:
        padded.write(">test_padded\n")
        for n in range(10):
            padded.write("N" * 80)
            padded.write("\n")
        padded.write("N" * 30)
        padded.write("A" * 20)
        padded.write("N" * 30)
        padded.write("\n")
        for n in range(10):
            padded.write("N" * 80)
            padded.write("\n")

    fasta = Fasta(filename)
    expect = 20
    result = fasta["test_padded"].unpadded_len
    print(expect, result)
    assert expect == result
    os.remove('data/padded.fasta')
    os.remove('data/padded.fasta.fai')

def test_numpy_array(remove_index):
    """ Test the __array_interface__ """
    import numpy
    filename = "data/genes.fasta.lower"
    reference = Fasta(filename)
    np_array = numpy.asarray(reference[0])
    assert isinstance(np_array, numpy.ndarray)

@pytest.fixture
def remove_index_mutable():
    with open('data/genes_mutable.fasta', 'wb') as mutable:
        mutable.write(open('data/genes.fasta', 'rb').read())
    mutable_fasta = Fasta('data/genes_mutable.fasta', mutable=True)
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file
    try:
        os.remove('data/genes_mutable.fasta')
    except EnvironmentError:
        pass  # some tests may delete this file
    try:
        os.remove('data/genes_mutable.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file

    def test_mutate_fasta_to_same(remove_index_mutable):
        mutable = Fasta('data/genes_mutable.fasta', mutable=True)
        fasta = Fasta('data/genes.fasta', mutable=False)
        chunk = fasta['gi|557361099|gb|KF435150.1|'][0:100]
        mutable['gi|557361099|gb|KF435150.1|'][0:100] = chunk.seq
        assert str(fasta['gi|557361099|gb|KF435150.1|']) == str(
            mutable['gi|557361099|gb|KF435150.1|'])

    def test_mutate_fasta_to_N(remove_index_mutable):
        mutable = Fasta('data/genes_mutable.fasta', mutable=True)
        chunk = 100 * 'N'
        mutable['gi|557361099|gb|KF435150.1|'][0:100] = chunk
        assert mutable['gi|557361099|gb|KF435150.1|'][0:100].seq == chunk

    def test_mutate_single_position(remove_index_mutable):
        mutable = Fasta('data/genes_mutable.fasta', mutable=True)
        chunk = 'N'
        mutable['gi|557361099|gb|KF435150.1|'][0] = chunk
        assert mutable['gi|557361099|gb|KF435150.1|'][0].seq == chunk

    @pytest.mark.xfail(raises=TypeError)
    def test_mutate_immutable_fasta(remove_index_mutable):
        mutable = Fasta('data/genes_mutable.fasta', mutable=False)
        chunk = 100 * 'N'
        mutable['gi|557361099|gb|KF435150.1|'][0:100] = chunk

    @pytest.mark.xfail(raises=IOError)
    def test_mutate_too_long(remove_index_mutable):
        mutable = Fasta('data/genes_mutable.fasta', mutable=True)
        chunk = 101 * 'N'
        mutable['gi|557361099|gb|KF435150.1|'][0:100] = chunk



import os
import pytest
from pyfaidx import Faidx, Fasta, FetchError

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def setup_zero():
    with open('data/zero_length.fasta', 'w') as fasta:
        fasta.write(""">A
ATCG
>B
>C

>D
GTA
GC""")
    yield
    os.remove('data/zero_length.fasta')
    os.remove('data/zero_length.fasta.fai')
              
def test_index_zero_length(setup_zero):
    fasta = Fasta('data/zero_length.fasta')
    
def test_fetch_zero_length(setup_zero):
    fasta = Fasta('data/zero_length.fasta')
    b = fasta["B"]
    assert str(b) == ''
        
@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file
        
def test_as_raw_zero_length_subsequence(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True, strict_bounds=True)
    expect = ''
    result = fasta['gi|557361099|gb|KF435150.1|'][100:100]
    assert result == expect

def test_zero_length_subsequence(remove_index):
    fasta = Fasta('data/genes.fasta', strict_bounds=True)
    expect = ''
    result = fasta['gi|557361099|gb|KF435150.1|'][100:100]
    assert result.seq == expect

def test_fetch_whole_entry(remove_index):
    faidx = Faidx('data/genes.fasta')
    expect = ('ATGACATCATTTTCCACCTCTGCTCAGTGTTCAACATCTGA'
            'CAGTGCTTGCAGGATCTCTCCTGGACAAATCAATCAGGTACGACCA'
            'AAACTGCCGCTTTTGAAGATTTTGCATGCAGCAGGTGCGCAAGG'
            'TGAAATGTTCACTGTTAAAGAGGTCATGCACTATTTAGGTCAGTACAT'
            'AATGGTGAAGCAACTTTATGATCAGCAGGAGCAGCATATGGTATATTG'
            'TGGTGGAGATCTTTTGGGAGAACTACTGGGACGTCAGAGCTTCTCCGTG'
            'AAAGACCCAAGCCCTCTCTATGATATGCTAAGAAAGAATCTTGTCACTTT'
            'AGCCACTGCTACTACAGCAAAGTGCAGAGGAAAGTTCCACTTCCAGAAAAA'
            'GAACTACAGAAGACGATATCCCCACACTGCCTACCTCAGAGCATAAATGCA'
            'TACATTCTAGAGAAGGTGATTGAAGTGGGAAAAAATGATGACCTGGAGGACTC')
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         1, 481)
    assert str(result) == expect

def test_fetch_middle(remove_index):
    faidx = Faidx('data/genes.fasta')
    expect = 'TTGAAGATTTTGCATGCAGCAGGTGCGCAAGGTGAAATGTTCACTGTTAAA'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         100, 150)
    assert str(result) == expect

def test_fetch_end(remove_index):
    faidx = Faidx('data/genes.fasta')
    expect = 'TC'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 481)
    assert str(result) == expect

def test_fetch_border(remove_index):
    """ Fetch past the end of a gene entry """
    faidx = Faidx('data/genes.fasta')
    expect = 'TC'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 500)
    assert str(result) == expect

def test_rev(remove_index):
    faidx = Faidx('data/genes.fasta')
    expect = 'GA'
    result = faidx.fetch('gi|557361099|gb|KF435150.1|',
                         480, 481)
    assert str(-result) == expect, result

def test_fetch_past_bounds(remove_index):
    """ Fetch past the end of a gene entry """
    faidx = Faidx('data/genes.fasta', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', 480, 5000)

def test_fetch_negative(remove_index):
    """ Fetch starting with a negative coordinate """
    faidx = Faidx('data/genes.fasta', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', -10, 10)

def test_fetch_reversed_coordinates(remove_index):
    """ Fetch starting with a negative coordinate """
    faidx = Faidx('data/genes.fasta', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|557361099|gb|KF435150.1|', 50, 10)

def test_fetch_keyerror(remove_index):
    """ Fetch a key that does not exist """
    faidx = Faidx('data/genes.fasta', strict_bounds=True)
    with pytest.raises(FetchError):
        result = faidx.fetch('gi|joe|gb|KF435150.1|', 1, 10)

def test_blank_string(remove_index):
    """ seq[0:0] should return a blank string mdshw5/pyfaidx#53 """
    fasta = Fasta('data/genes.fasta', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][0:0] == ''

def test_slice_from_beginning(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][:4] == 'ATGA'

def test_slice_from_end(remove_index):
    fasta = Fasta('data/genes.fasta', as_raw=True)
    assert fasta['gi|557361099|gb|KF435150.1|'][-4:] == 'ACTC'

def test_issue_74_start(remove_index):
    f0 = Fasta('data/genes.fasta', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta', one_based_attributes=True)
    assert f0['gi|557361099|gb|KF435150.1|'][0:90].start == f1['gi|557361099|gb|KF435150.1|'][0:90].start - 1

def test_issue_74_consistency(remove_index):
    f0 = Fasta('data/genes.fasta', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta', one_based_attributes=True)
    assert str(f0['gi|557361099|gb|KF435150.1|'][0:90]) == str(f1['gi|557361099|gb|KF435150.1|'][0:90])

def test_issue_74_end_faidx(remove_index):
    f0 = Faidx('data/genes.fasta', one_based_attributes=False)
    f1 = Faidx('data/genes.fasta', one_based_attributes=True)
    end0 = f0.fetch('gi|557361099|gb|KF435150.1|', 1, 90).end
    end1 = f1.fetch('gi|557361099|gb|KF435150.1|', 1, 90).end
    assert end0 == end1

def test_issue_74_end_fasta(remove_index):
    f0 = Fasta('data/genes.fasta', one_based_attributes=False)
    f1 = Fasta('data/genes.fasta', one_based_attributes=True)
    end0 = f0['gi|557361099|gb|KF435150.1|'][1:90].end
    end1 = f1['gi|557361099|gb|KF435150.1|'][1:90].end
    print((end0, end1))
    assert end0 == end1

def test_issue_79_fix(remove_index):
    f = Fasta('data/genes.fasta')
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print((s.start, s.end))
    assert (101, 105) == (s.start, s.end)

def test_issue_79_fix_negate(remove_index):
    f = Fasta('data/genes.fasta')
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    s = -s
    print((s.start, s.end))
    assert (105, 101) == (s.start, s.end)

def test_issue_79_fix_one_based_false(remove_index):
    f = Fasta('data/genes.fasta', one_based_attributes=False)
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print((s.start, s.end))
    assert (100, 105) == (s.start, s.end)

def test_issue_79_fix_one_based_false_negate(remove_index):
    f = Fasta('data/genes.fasta', one_based_attributes=False)
    s = f['gi|557361099|gb|KF435150.1|'][100:105]
    print(s.__dict__)
    s = -s
    print(s.__dict__)
    assert (105, 100) == (s.start, s.end)


import os
import pytest
from pyfaidx import Fasta

path = os.path.dirname(__file__)
os.chdir(path)

@pytest.fixture
def remove_index():
    yield
    try:
        os.remove('data/genes.fasta.fai')
    except EnvironmentError:
        pass  # some tests may delete this file
    
def test_integer_slice(remove_index):
    fasta = Fasta('data/genes.fasta')
    expect = fasta['gi|563317589|dbj|AB821309.1|'][:100].seq
    result = fasta[0][:100].seq
    assert expect == result

def test_integer_index(remove_index):
    fasta = Fasta('data/genes.fasta')
    expect = fasta['gi|563317589|dbj|AB821309.1|'][100].seq
    result = fasta[0][100].seq
    assert expect == result



import os
import pytest
try:
    from collections import OrderedDict
except ImportError: #python 2.6
    from ordereddict import OrderedDict
import threading
from pyfaidx import Fasta
import random
import tempfile
import time
import shutil
from unittest import TestCase

path = os.path.dirname(__file__)
os.chdir(path)


class _ThreadReadSequence(threading.Thread):
    def __init__(self, rand, result_map, result_lock, name, seq):
        super(_ThreadReadSequence, self).__init__()

        seq_len = len(seq)
        sub_seq_slices = list(slice(i, min(i + 20, seq_len)) for i in range(0, seq_len, 20))
        random.shuffle(sub_seq_slices)

        self.result_map = result_map
        self.result_lock = result_lock
        self.name = name
        self.seq = seq
        self.sub_seq_slices = sub_seq_slices

    def run(self):
        name = self.name
        seq = self.seq

        sub_seqs = [''] * len(self.sub_seq_slices)
        for sub_seq_slice in self.sub_seq_slices:
            sub_seqs[sub_seq_slice.start//20] = seq[sub_seq_slice]
            time.sleep(0)

        # Put sub-sequences in correct order
        seq_str = ''.join(sub_seqs)

        with self.result_lock:
            self.result_map[name] = seq_str


class _ThreadWriteSequence(threading.Thread):
    def __init__(self, rand, name, seq):
        super(_ThreadWriteSequence, self).__init__()

        seq_len = len(seq)
        sub_seq_slices = list(slice(i, min(i + 20, seq_len)) for i in range(0, seq_len, 20))
        random.shuffle(sub_seq_slices)

        self.name = name
        self.seq = seq
        self.sub_seq_slices = sub_seq_slices

    def run(self):
        seq = self.seq
        seq_len = len(seq)
        seq_str = seq[:].lower()

        for sub_seq_slice in self.sub_seq_slices:
            try:
                seq[sub_seq_slice] = seq_str[sub_seq_slice]
                time.sleep(0)
            except Exception:
                # Conflicting simultaneous writes are likely to cause exceptions
                # We test for the expected string at the end, so ignore interim
                # failures.
                pass


class TestFastaIntIndex(TestCase):
    def setUp(self):
        self.longMessage = True
        self.maxDiff = None
        self.tmp_dir = tempfile.mkdtemp()
        # Use a seeded random orders are randomish within a test, but the same across test runs
        self.rand = random.Random(8903423147)

    def tearDown(self):
        tmp_dir = getattr(self, 'tmp_dir', None)
        if tmp_dir:
            shutil.rmtree(tmp_dir)

        try:
            os.remove('data/genes.fasta.fai')
        except EnvironmentError:
            pass  # some tests may delete this file

    def test_simultaneous_reads(self):
        """
        Test that each read of a sequence range is atomic.
        To do this, spawn several threads to simultaneously read the sequences
        in a Fasta file in pieces. If the reads are not atomic, then it is
        reasonably likely (with sufficient concurrency) that a read from one
        thread will affect that in another, so the sequences will not be
        read properly.
        """
        # Read in original file data
        ref_seq_map = OrderedDict()
        with Fasta('data/genes.fasta', as_raw=True, strict_bounds=True) as fasta:
            for name, seq in fasta.records.items():
                ref_seq_map[name] = seq[:]

        # Initialize map with fasta sequence names to enforce same ordering as 'ref_seq_map'
        thread_result_lock = threading.Lock()
        thread_read_seq_map = OrderedDict((name, None) for name in ref_seq_map)

        # Read file again, using many threads and simultaneously reading each sequence in pieces
        with Fasta('data/genes.fasta', as_raw=True, strict_bounds=True) as fasta:
            threads = []
            for name, seq in fasta.records.items():
                threads.append(_ThreadReadSequence(self.rand, thread_read_seq_map, thread_result_lock, name, seq))

            for thread in threads:
                thread.start()

            for thread in threads:
                thread.join()

        self.assertEqual(thread_read_seq_map, ref_seq_map)

    def test_simultaneous_writes(self):
        """
        Test that each write of a sequence range is atomic.
        To do this, spawn several threads to simultaneously write sequences
        to a Fasta file in pieces. If the writes are not atomic, then it is
        reasonably likely (with sufficient concurrency) that a write from one
        thread will affect that in another, so the sequences will not be
        written properly. To make sure all sequences are mutated, the writes
        will transform the sequence to lower-case.
        """

        tmp_dir = self.tmp_dir

        tmp_fasta = os.path.join(tmp_dir, 'genes_write.fasta')
        shutil.copyfile('data/genes.fasta', tmp_fasta)

        # Read in original file data
        ref_seq_map = OrderedDict()
        with Fasta('data/genes.fasta', as_raw=True, strict_bounds=True) as fasta:
            for name, seq in fasta.records.items():
                ref_seq_map[name] = seq[:].lower()

        # Now write file, using many threads and simultaneously reading each sequence in pieces
        with Fasta(tmp_fasta, as_raw=True, strict_bounds=True, mutable=True) as fasta:
            threads = []
            for name, seq in fasta.records.items():
                threads.append(_ThreadWriteSequence(self.rand, name, seq))

            for thread in threads:
                thread.start()

            for thread in threads:
                thread.join()

            fasta.faidx.file.flush()

        # Now read written Fasta file, and compare it to the original
        thread_write_seq_map = OrderedDict()
        with Fasta(tmp_fasta, as_raw=True, strict_bounds=True) as fasta:
            for name, seq in fasta.records.items():
                thread_write_seq_map[name] = seq[:]

        self.assertEqual(thread_write_seq_map, ref_seq_map)

    def test_simultaneous_reads_and_writes(self):
        """
        Combine the above two tests to check that interleaved reads and writes don't conflict.
        """

        tmp_dir = self.tmp_dir

        tmp_fasta = os.path.join(tmp_dir, 'genes_write.fasta')
        shutil.copyfile('data/genes.fasta', tmp_fasta)

        # Read in original file data
        ref_seq_map = OrderedDict()
        with Fasta('data/genes.fasta', as_raw=True, strict_bounds=True) as fasta:
            for name, seq in fasta.records.items():
                ref_seq_map[name] = seq[:].lower()

        # Initialize map with fasta sequence names to enforce same ordering as 'ref_seq_map'
        thread_result_lock = threading.Lock()
        thread_read_seq_map = OrderedDict((name, None) for name in ref_seq_map)

        # Now write file, using many threads and simultaneously reading each sequence in pieces
        with Fasta(tmp_fasta, as_raw=True, strict_bounds=True, mutable=True) as fasta:
            threads = []
            for name, seq in fasta.records.items():
                threads.append(_ThreadWriteSequence(self.rand, name, seq))
                threads.append(_ThreadReadSequence(self.rand, thread_read_seq_map, thread_result_lock, name, seq))

            for thread in threads:
                thread.start()

            for thread in threads:
                thread.join()

            fasta.faidx.file.flush()

        # Now read written Fasta file, and compare it to the original
        thread_write_seq_map = OrderedDict()
        with Fasta(tmp_fasta, as_raw=True, strict_bounds=True) as fasta:
            for name, seq in fasta.records.items():
                thread_write_seq_map[name] = seq[:]

        # Change read strings to lower-case (may be a mixture of lower and upper)
        for name in ref_seq_map.keys():
            thread_read_seq_map[name] = thread_read_seq_map[name].lower()

        self.assertEqual(thread_write_seq_map, ref_seq_map)
        self.assertEqual(thread_read_seq_map, ref_seq_map)

